@@ -1,4 +1,20 @@
-/*
+/*  Copyright 1986-1992 Emmet P. Gray.
+ *  Copyright 1994,1996-2009 Alain Knaff.
+ *  This file is part of mtools.
+ *
+ *  Mtools is free software: you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
+ *
+ *  Mtools is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with Mtools.  If not, see <http://www.gnu.org/licenses/>.
+ *
  * mformat.c
  */
 
@@ -21,6 +37,7 @@
 #include "xdf_io.h"
 #endif
 #include "partition.h"
+#include "file_name.h"
 
 #ifndef abs
 #define abs(x) ((x)>0?(x):-(x))
@@ -37,9 +54,7 @@
 #endif
 
 
-extern int errno;
-
-static int init_geometry_boot(struct bootsector *boot, struct device *dev,
+static int init_geometry_boot(union bootsector *boot, struct device *dev,
 			       int sectors0, int rate_0, int rate_any,
 			       unsigned long *tot_sectors, int keepBoot)
 {
@@ -50,45 +65,45 @@ static int init_geometry_boot(struct bootsector *boot, struct device *dev,
 	int j;
 	int sum;
 
-	set_word(boot->nsect, dev->sectors);
-	set_word(boot->nheads, dev->heads);
+	set_word(boot->boot.nsect, dev->sectors);
+	set_word(boot->boot.nheads, dev->heads);
 
 	*tot_sectors = dev->heads * dev->sectors * dev->tracks - DWORD(nhs);
 
 	if (*tot_sectors < 0x10000){
-		set_word(boot->psect, *tot_sectors);
-		set_dword(boot->bigsect, 0);
+		set_word(boot->boot.psect, *tot_sectors);
+		set_dword(boot->boot.bigsect, 0);
 	} else {
-		set_word(boot->psect, 0);
-		set_dword(boot->bigsect, *tot_sectors);
+		set_word(boot->boot.psect, 0);
+		set_dword(boot->boot.bigsect, *tot_sectors);
 	}
 
 	if (dev->use_2m & 0x7f){
 		int bootOffset;
-		strncpy(boot->banner, "2M-STV04", 8);
-		boot->ext.old.res_2m = 0;
-		boot->ext.old.fmt_2mf = 6;
+		strncpy(boot->boot.banner, "2M-STV04", 8);
+		boot->boot.ext.old.res_2m = 0;
+		boot->boot.ext.old.fmt_2mf = 6;
 		if ( dev->sectors % ( ((1 << dev->ssize) + 3) >> 2 ))
-			boot->ext.old.wt = 1;
+			boot->boot.ext.old.wt = 1;
 		else
-			boot->ext.old.wt = 0;
-		boot->ext.old.rate_0= rate_0;
-		boot->ext.old.rate_any= rate_any;
-		if (boot->ext.old.rate_any== 2 )
-			boot->ext.old.rate_any= 1;
+			boot->boot.ext.old.wt = 0;
+		boot->boot.ext.old.rate_0= rate_0;
+		boot->boot.ext.old.rate_any= rate_any;
+		if (boot->boot.ext.old.rate_any== 2 )
+			boot->boot.ext.old.rate_any= 1;
 		i=76;
 
 		/* Infp0 */
-		set_word(boot->ext.old.Infp0, i);
-		boot->jump[i++] = sectors0;
-		boot->jump[i++] = 108;
+		set_word(boot->boot.ext.old.Infp0, i);
+		boot->bytes[i++] = sectors0;
+		boot->bytes[i++] = 108;
 		for(j=1; j<= sectors0; j++)
-			boot->jump[i++] = j;
+			boot->bytes[i++] = j;
 
-		set_word(boot->ext.old.InfpX, i);
+		set_word(boot->boot.ext.old.InfpX, i);
 		
-		boot->jump[i++] = 64;
-		boot->jump[i++] = 3;
+		boot->bytes[i++] = 64;
+		boot->bytes[i++] = 3;
 		nb_renum = i++;
 		sector2 = dev->sectors;
 		size2 = dev->ssize;
@@ -96,38 +111,38 @@ static int init_geometry_boot(struct bootsector *boot, struct device *dev,
 		while( sector2 ){
 			while ( sector2 < (1 << size2) >> 2 )
 				size2--;
-			boot->jump[i++] = 128 + j;
-			boot->jump[i++] = j++;
-			boot->jump[i++] = size2;
+			boot->bytes[i++] = 128 + j;
+			boot->bytes[i++] = j++;
+			boot->bytes[i++] = size2;
 			sector2 -= (1 << size2) >> 2;
 		}
-		boot->jump[nb_renum] = ( i - nb_renum - 1 ) / 3;
+		boot->bytes[nb_renum] = ( i - nb_renum - 1 ) / 3;
 
-		set_word(boot->ext.old.InfTm, i);
+		set_word(boot->boot.ext.old.InfTm, i);
 
 		sector2 = dev->sectors;
 		size2= dev->ssize;
 		while(sector2){
 			while ( sector2 < 1 << ( size2 - 2) )
 				size2--;
-			boot->jump[i++] = size2;
+			boot->bytes[i++] = size2;
 			sector2 -= 1 << (size2 - 2 );
 		}
 		
-		set_word(boot->ext.old.BootP,i);
+		set_word(boot->boot.ext.old.BootP,i);
 		bootOffset = i;
 
 		/* checksum */		
 		for (sum=0, j=64; j<i; j++) 
-			sum += boot->jump[j];/* checksum */
-		boot->ext.old.CheckSum=-sum;
+			sum += boot->bytes[j];/* checksum */
+		boot->boot.ext.old.CheckSum=-sum;
 		return bootOffset;
 	} else {
 		if(!keepBoot) {
-			boot->jump[0] = 0xeb;
-			boot->jump[1] = 0;
-			boot->jump[2] = 0x90;
-			strncpy(boot->banner, "MTOOL399", 8);
+			boot->boot.jump[0] = 0xeb;
+			boot->boot.jump[1] = 0;
+			boot->boot.jump[2] = 0x90;
+			strncpy(boot->boot.banner, mformat_banner, 8);
 			/* It looks like some versions of DOS are
 			 * rather picky about this, and assume default
 			 * parameters without this, ignoring any
@@ -243,7 +258,7 @@ static void calc_fat_bits2(Fs_t *Fs, unsigned long tot_sectors, int fat_bits,
 				   <= MAX_BYTES_PER_CLUSTER)
 					Fs->cluster_size <<= 1;
 				else if(may_change_root_size) {
-					Fs->dir_len += 
+					Fs->dir_len +=
 						rem_sect - MY_DISK_SIZE(12, FAT12-1);
 				}
 				set_fat12(Fs);
@@ -263,7 +278,7 @@ static void calc_fat_bits2(Fs_t *Fs, unsigned long tot_sectors, int fat_bits,
 	}
 }
 
-static __inline__ void format_root(Fs_t *Fs, char *label, struct bootsector *boot)
+static __inline__ void format_root(Fs_t *Fs, char *label, union bootsector *boot)
 {
 	Stream_t *RootDir;
 	char *buf;
@@ -290,9 +305,9 @@ static __inline__ void format_root(Fs_t *Fs, char *label, struct bootsector *boo
 		dirlen = Fs->cluster_size;
 		fatAllocate(Fs, Fs->rootCluster, Fs->end_fat);
 	} else
-		dirlen = Fs->dir_len; 
+		dirlen = Fs->dir_len;
 	for (i = 0; i < dirlen; i++)
-		WRITES(RootDir, buf, sectorsToBytes((Stream_t*)Fs, i),  
+		WRITES(RootDir, buf, sectorsToBytes((Stream_t*)Fs, i),
 			   Fs->sector_size);
 
 	ch.ignore_entry = 1;
@@ -301,14 +316,14 @@ static __inline__ void format_root(Fs_t *Fs, char *label, struct bootsector *boo
 
 	FREE(&RootDir);
 	if(Fs->fat_bits == 32)
-		set_word(boot->dirents, 0);
+		set_word(boot->boot.dirents, 0);
 	else
-		set_word(boot->dirents, Fs->dir_len * (Fs->sector_size / 32));
+		set_word(boot->boot.dirents, Fs->dir_len * (Fs->sector_size / 32));
 	free(buf);
 }
 
 
-static void xdf_calc_fat_size(Fs_t *Fs, unsigned long tot_sectors, 
+static void xdf_calc_fat_size(Fs_t *Fs, unsigned long tot_sectors,
 			      int fat_bits)
 {
 	unsigned int rem_sect;
@@ -346,7 +361,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 	unsigned int slack;
 	int printGrowMsg=1; /* Should we print "growing FAT" messages ?*/
 	
-#if DEBUG
+#ifdef DEBUG
 	fprintf(stderr, "Fat start=%d\n", Fs->fat_start);
 	fprintf(stderr, "tot_sectors=%lu\n", tot_sectors);
 	fprintf(stderr, "dir_len=%d\n", Fs->dir_len);
@@ -361,7 +376,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 	   Fs->cluster_size %2 == 0)
 		rem_sect--;
 
-#if DEBUG
+#ifdef DEBUG
 	fprintf(stderr, "Rem sect=%lu\n", rem_sect);
 #endif
 
@@ -372,13 +387,13 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 
 
 	/* See fat_size_calculation.tex or
-	   (http://www.mtools.linux.lu/fat_size_calculation.pdf) for an
-	   explantation about why the stuff below works...
+	   (http://ftp.gnu.org/software/mtools/manual/fat_size_calculation.pdf)
+	   for an explantation about why the stuff below works...
 	*/
 
 	fat_nybbles = Fs->fat_bits / 4;
 	numerator   = rem_sect+2*Fs->cluster_size;
-	denominator = 
+	denominator =
 	  Fs->cluster_size * Fs->sector_size * 2 +
 	  Fs->num_fat * fat_nybbles;
 
@@ -389,7 +404,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 		 * rather than multiplying the numerator */
 		denominator = denominator / fat_nybbles;
 
-#if DEBUG
+#ifdef DEBUG
 	fprintf(stderr, "Numerator=%lu denominator=%lu\n",
 		numerator, denominator);
 #endif
@@ -404,16 +419,16 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 		Fs->num_clus = FAT12-1;
 	
 	/* A safety, if above math is correct, this should not be happen...*/
-	if(Fs->num_clus > (Fs->fat_len * Fs->sector_size * 2 / 
+	if(Fs->num_clus > (Fs->fat_len * Fs->sector_size * 2 /
 			   fat_nybbles - 2)) {
-		fprintf(stderr, 
+		fprintf(stderr,
 			"Fat size miscalculation, shrinking num_clus from %d ",
 			Fs->num_clus);
-		Fs->num_clus = (Fs->fat_len * Fs->sector_size * 2 / 
+		Fs->num_clus = (Fs->fat_len * Fs->sector_size * 2 /
 				fat_nybbles - 2);
 		fprintf(stderr, " to %d\n", Fs->num_clus);
 	}
-#if DEBUG
+#ifdef DEBUG
 	fprintf(stderr, "Num_clus=%d fat_len=%d nybbles=%d\n",
 		Fs->num_clus, Fs->fat_len, fat_nybbles);
 #endif
@@ -431,11 +446,11 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 	 * above, we will have to grow the FAT in order to take up any excess
 	 * sectors... */
 #ifdef HAVE_ASSERT_H
-	assert(rem_sect >= Fs->num_clus * Fs->cluster_size + 
+	assert(rem_sect >= Fs->num_clus * Fs->cluster_size +
 	       Fs->fat_len * Fs->num_fat);
 #endif
-	slack = rem_sect - 
-		Fs->num_clus * Fs->cluster_size - 
+	slack = rem_sect -
+		Fs->num_clus * Fs->cluster_size -
 		Fs->fat_len * Fs->num_fat;
 	if(slack >= Fs->cluster_size) {
 		/* This can happen under two circumstances:
@@ -444,10 +459,10 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 		*/
 		if(printGrowMsg) {
 			fprintf(stderr, "Slack=%d\n", slack);
-			fprintf(stderr, "Growing fat size from %d", 
+			fprintf(stderr, "Growing fat size from %d",
 				Fs->fat_len);
 		}
-		Fs->fat_len += 
+		Fs->fat_len +=
 			(slack - Fs->cluster_size) / Fs->num_fat + 1;
 		if(printGrowMsg) {
 			fprintf(stderr,
@@ -461,7 +476,7 @@ static void calc_fat_size(Fs_t *Fs, unsigned long tot_sectors)
 
 #ifdef HAVE_ASSERT_H
 	/* Fat must be big enough for all clusters */
-	assert( ((Fs->num_clus+2) * fat_nybbles) <= 
+	assert( ((Fs->num_clus+2) * fat_nybbles) <=
 		(Fs->fat_len*Fs->sector_size*2));
 
 	/* num_clus must be big enough to cover rest of disk, or else further
@@ -480,27 +495,27 @@ static unsigned char bootprog[]=
  0xb9, 0x01, 0x00, 0xcd, 0x13, 0x72, 0x05, 0xea, 0x00, 0x7c, 0x00,
  0x00, 0xcd, 0x19};
 
-static __inline__ void inst_boot_prg(struct bootsector *boot, int offset)
+static __inline__ void inst_boot_prg(union bootsector *boot, int offset)
 {
-	memcpy((char *) boot->jump + offset, 
+	memcpy((char *) boot->boot.jump + offset,
 	       (char *) bootprog, sizeof(bootprog) /sizeof(bootprog[0]));
 	if(offset - 2 < 0x80) {
 	  /* short jump */
-	  boot->jump[0] = 0xeb;
-	  boot->jump[1] = offset -2;
-	  boot->jump[2] = 0x90;
+	  boot->boot.jump[0] = 0xeb;
+	  boot->boot.jump[1] = offset -2;
+	  boot->boot.jump[2] = 0x90;
 	} else {
 	  /* long jump, if offset is too large */
-	  boot->jump[0] = 0xe9;
-	  boot->jump[1] = offset -3;
-	  boot->jump[2] = 0x00;
+	  boot->boot.jump[0] = 0xe9;
+	  boot->boot.jump[1] = offset -3;
+	  boot->boot.jump[2] = 0x00;
 	}
-	set_word(boot->jump + offset + 20, offset + 24);
+	set_word(boot->boot.jump + offset + 20, offset + 24);
 }
 
 static void calc_cluster_size(struct Fs_t *Fs, unsigned long tot_sectors,
 			      int fat_bits)
-			      
+			
 {
 	unsigned int max_clusters; /* maximal possible number of sectors for
 				   * this FAT entry length (12/16/32) */
@@ -512,16 +527,16 @@ static void calc_cluster_size(struct Fs_t *Fs, unsigned long tot_sectors,
 	switch(abs(fat_bits)) {
 		case 12:			
 			max_clusters = FAT12-1;
-			max_fat_size = Fs->num_fat * 
+			max_fat_size = Fs->num_fat *
 				FAT_SIZE(12, Fs->sector_size, max_clusters);
 			break;
 		case 16:
 		case 0: /* still hesititating between 12 and 16 */
 			max_clusters = FAT16-1;
-			max_fat_size = Fs->num_fat * 
+			max_fat_size = Fs->num_fat *
 				FAT_SIZE(16, Fs->sector_size, max_clusters);
 			break;
-		case 32:		  
+		case 32:		
 			Fs->cluster_size = 8;
 			/* According to
 			 * http://support.microsoft.com/support/kb/articles/q154/9/97.asp
@@ -566,8 +581,8 @@ static int old_dos_size_to_geom(size_t size, int *cyls, int *heads, int *sects)
 	unsigned int i;
 	size = size * 2;
 	for(i=0; i < sizeof(old_dos) / sizeof(old_dos[0]); i++){
-		if (old_dos[i].sectors * 
-		    old_dos[i].tracks * 
+		if (old_dos[i].sectors *
+		    old_dos[i].tracks *
 		    old_dos[i].heads == size) {
 			*cyls = old_dos[i].tracks;
 			*heads = old_dos[i].heads;
@@ -580,7 +595,7 @@ static int old_dos_size_to_geom(size_t size, int *cyls, int *heads, int *sects)
 
 
 static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
-			       struct Fs_t *Fs, struct bootsector *boot)
+			       struct Fs_t *Fs, union bootsector *boot)
 {
 	unsigned int i;
 
@@ -590,9 +605,9 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 		    dev->heads == old_dos[i].heads &&
 		    (dev->fat_bits == 0 || abs(dev->fat_bits) == 12) &&
 		    (Fs->dir_len == 0 || Fs->dir_len == old_dos[i].dir_len) &&
-		    (Fs->cluster_size == 0 || 
+		    (Fs->cluster_size == 0 ||
 		     Fs->cluster_size == old_dos[i].cluster_size)) {
-			boot->descr = old_dos[i].media;
+			boot->boot.descr = old_dos[i].media;
 			Fs->cluster_size = old_dos[i].cluster_size;
 			Fs->dir_len = old_dos[i].dir_len;
 			Fs->fat_len = old_dos[i].fat_len;
@@ -606,9 +621,9 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 
 		/* a non-standard format */
 		if(DWORD(nhs))
-			boot->descr = 0xf8;
+			boot->boot.descr = 0xf8;
 		  else
-			boot->descr = 0xf0;
+			boot->boot.descr = 0xf0;
 
 
 		if(!Fs->cluster_size) {
@@ -639,18 +654,18 @@ static void calc_fs_parameters(struct device *dev, unsigned long tot_sectors,
 		}
 	}
 
-	set_word(boot->fatlen, Fs->fat_len);
+	set_word(boot->boot.fatlen, Fs->fat_len);
 }
 
 
 
 static void calc_fs_parameters_32(unsigned long tot_sectors,
-				  struct Fs_t *Fs, struct bootsector *boot)
+				  struct Fs_t *Fs, union bootsector *boot)
 {
 	if(DWORD(nhs))
-		boot->descr = 0xf8;
+		boot->boot.descr = 0xf8;
 	else
-		boot->descr = 0xf0;
+		boot->boot.descr = 0xf0;
 	if(!Fs->cluster_size)
 		/* According to
 		 * http://www.microsoft.com/kb/articles/q154/9/97.htm,
@@ -662,18 +677,18 @@ static void calc_fs_parameters_32(unsigned long tot_sectors,
 	Fs->num_clus = tot_sectors / Fs->cluster_size;
 	set_fat32(Fs);
 	calc_fat_size(Fs, tot_sectors);
-	set_word(boot->fatlen, 0);
-	set_dword(boot->ext.fat32.bigFat, Fs->fat_len);
+	set_word(boot->boot.fatlen, 0);
+	set_dword(boot->boot.ext.fat32.bigFat, Fs->fat_len);
 }
 
 
 
 
-static void usage(void)
+static void usage(int ret)
 {
-	fprintf(stderr, 
+	fprintf(stderr,
 		"Mtools version %s, dated %s\n", mversion, mdate);
-	fprintf(stderr, 
+	fprintf(stderr,
 		"Usage: %s [-V] [-t tracks] [-h heads] [-n sectors] "
 		"[-v label] [-1] [-4] [-8] [-f size] "
 		"[-N serialnumber] "
@@ -686,9 +701,56 @@ static void usage(void)
 		"[-S hardsectorsize] [-M softsectorsize] [-3] "
 		"[-2 track0sectors] [-0 rate0] [-A rateany] [-a]"
 		"device\n", progname);
-	exit(1);
+	exit(ret);
 }
 
+#ifdef OS_linux
+static int get_block_geom(int fd, struct MT_STAT *buf, struct device *dev,
+			  char *errmsg) {
+	struct hd_geometry geom;
+	long size;
+	int heads=dev->heads;
+	int sectors=dev->sectors;
+	int sect_per_track;
+
+	if (ioctl(fd, HDIO_GETGEO, &geom) < 0) {
+		sprintf(errmsg, "Could not get geometry of device (%s)",
+			strerror(errno));
+		return -1;
+	}
+	
+	if (ioctl(fd, BLKGETSIZE, &size) < 0) {
+		sprintf(errmsg, "Could not get size of device (%s)",
+			strerror(errno));
+		return -1;
+	}
+	
+	if(!heads)
+		heads = geom.heads;
+	if(!sectors)
+		sectors = geom.sectors;
+
+	sect_per_track = heads * sectors;
+	if(!dev->hidden) {
+		int hidden;
+		hidden = geom.start % sect_per_track;
+		if(hidden && hidden != sectors) {
+			sprintf(errmsg,
+				"Hidden (%d) does not match sectors (%d)\n",
+				hidden, sectors);
+			return -1;
+		}
+		dev->hidden = hidden;
+	}
+	dev->heads = heads;
+	dev->sectors = sectors;
+	if(!dev->tracks)
+		dev->tracks = (size + dev->hidden) / sect_per_track;
+	size = dev->tracks * dev->heads * dev->sectors + dev->hidden;
+	return 0;
+}
+#endif
+
 void mformat(int argc, char **argv, int dummy)
 {
 	int r; /* generic return value */
@@ -710,7 +772,7 @@ void mformat(int argc, char **argv, int dummy)
 	int format_xdf = 0;
 	struct xdf_info info;
 #endif
-	struct bootsector *boot;
+	union bootsector boot;
 	char *bootSector=0;
 	int c;
 	int keepBoot = 0;
@@ -721,7 +783,9 @@ void mformat(int argc, char **argv, int dummy)
 
 	char drive, name[EXPAND_BUF];
 
-	char label[VBUFSIZE], buf[MAX_SECTOR], shortlabel[13];
+	char label[VBUFSIZE];
+
+	dos_name_t shortlabel;
 	struct device *dev;
 	char errmsg[200];
 
@@ -730,10 +794,10 @@ void mformat(int argc, char **argv, int dummy)
 	int fsVersion;
 	int mediaDesc=-1;
 
-	mt_off_t maxSize;
+	mt_size_t maxSize;
 
 	int Atari = 0; /* should we add an Atari-style serial number ? */
- 
+
 	hs = hs_set = 0;
 	argtracks = 0;
 	argheads = 0;
@@ -765,9 +829,11 @@ void mformat(int argc, char **argv, int dummy)
 	rate_any = mtools_rate_any;
 
 	/* get command line options */
+	if(helpFlag(argc, argv))
+		usage(0);
 	while ((c = getopt(argc,argv,
 			   "i:148f:t:n:v:qub"
-			   "kB:r:L:IFCc:Xh:s:l:N:H:M:S:2:30:Aad:m:"))!= EOF) {
+			   "kB:r:L:I:FCc:Xh:s:l:N:H:M:S:2:30:Aad:m:"))!= EOF) {
 		switch (c) {
 			case 'i':
 				set_cmd_line_image(optarg, 0);
@@ -790,7 +856,7 @@ void mformat(int argc, char **argv, int dummy)
 						       &argtracks, &argheads,
 						       &argsectors);
 				if(r) {
-					fprintf(stderr, 
+					fprintf(stderr,
 						"Bad size %s\n", optarg);
 					exit(1);
 				}
@@ -815,7 +881,7 @@ void mformat(int argc, char **argv, int dummy)
 			case 'u':
 			case 'b':
 			/*case 's': leave this for compatibility */
-				fprintf(stderr, 
+				fprintf(stderr,
 					"Flag %c not supported by mtools\n",c);
 				exit(1);
 				
@@ -830,7 +896,11 @@ void mformat(int argc, char **argv, int dummy)
 			case 'S':
 				argssize = atoi(optarg) | 0x80;
 				if(argssize < 0x81)
-					usage();
+					usage(1);
+				if(argssize >= 0x87) {
+					fprintf(stderr, "argssize must be less than 6\n");
+					usage(1);
+				}
 				break;
 
 #ifdef USE_XDF
@@ -861,7 +931,7 @@ void mformat(int argc, char **argv, int dummy)
 				   msize != 2048 &&
 				   msize != 4096) {
 				  fprintf(stderr, "Only sector sizes of 512, 1024, 2048 or 4096 bytes are allowed\n");
-				  usage();
+				  usage(1);
 				}
 				break;
 
@@ -890,7 +960,7 @@ void mformat(int argc, char **argv, int dummy)
 				Fs.cluster_size = atoi(optarg);
 				break;
 
-			case 'r': 
+			case 'r':
 				Fs.dir_len = strtoul(optarg,0,0);
 				break;
 			case 'L':
@@ -914,13 +984,13 @@ void mformat(int argc, char **argv, int dummy)
 				mediaDesc = strtoul(optarg,0,0);
 				break;
 			default:
-				usage();
+				usage(1);
 		}
 	}
 
 	if (argc - optind != 1 ||
 	    !argv[optind][0] || argv[optind][1] != ':')
-		usage();
+		usage(1);
 
 #ifdef USE_XDF
 	if(create && format_xdf) {
@@ -960,16 +1030,18 @@ void mformat(int argc, char **argv, int dummy)
 #endif
 			Fs.Direct = 0;
 #ifdef USE_FLOPPYD
-			Fs.Direct = FloppydOpen(&used_dev, dev, name, O_RDWR | create,
-									errmsg, 0, 1);
+			Fs.Direct = FloppydOpen(&used_dev, dev, name,
+						O_RDWR | create,
+						errmsg, 0, 1);
 			if(Fs.Direct) {
 				maxSize = max_off_t_31;
 			}
 #endif
 			if(!Fs.Direct) {			
 				Fs.Direct = SimpleFileOpen(&used_dev, dev, name,
-										   O_RDWR | create,
-										   errmsg, 0, 1, &maxSize);
+							   O_RDWR | create,
+							   errmsg, 0, 1,
+							   &maxSize);
 			}
 #ifdef USE_XDF
 		} else {
@@ -990,42 +1062,22 @@ void mformat(int argc, char **argv, int dummy)
 		if ((!used_dev.tracks || !used_dev.heads || !used_dev.sectors) &&
 			(!IS_SCSI(dev))) {
 			int fd= get_fd(Fs.Direct);
-			struct MT_STAT buf;
+			struct MT_STAT stbuf;
 
-			if (MT_FSTAT(fd, &buf) < 0) {
+			if (MT_FSTAT(fd, &stbuf) < 0) {
 				sprintf(errmsg, "Could not stat file (%s)", strerror(errno));
 				continue;						
 			}
 
-			if (S_ISBLK(buf.st_mode)) {
-				struct hd_geometry geom;
-				long size;
-				int sect_per_track;
-
-				if (ioctl(fd, HDIO_GETGEO, &geom) < 0) {
-					sprintf(errmsg, "Could not get geometry of device (%s)",
-							strerror(errno));
-					continue;
-				}
-
-				if (ioctl(fd, BLKGETSIZE, &size) < 0) {
-					sprintf(errmsg, "Could not get size of device (%s)",
-							strerror(errno));
-					continue;
-				}
-
-				sect_per_track = geom.heads * geom.sectors;
-				used_dev.heads = geom.heads;
-				used_dev.sectors = geom.sectors;
-				used_dev.hidden = geom.start % sect_per_track;
-				used_dev.tracks = (size + used_dev.hidden) / sect_per_track;
-			}
+			if (S_ISBLK(stbuf.st_mode) &&
+			    get_block_geom(fd, &stbuf, &used_dev, errmsg) < 0)
+				continue;
 		}
 #endif
 
 		/* no way to find out geometry */
 		if (!used_dev.tracks || !used_dev.heads || !used_dev.sectors){
-			sprintf(errmsg, 
+			sprintf(errmsg,
 				"Unknown geometry "
 				"(You must tell the complete geometry "
 				"of the disk, \neither in /etc/mtools.conf or "
@@ -1036,7 +1088,7 @@ void mformat(int argc, char **argv, int dummy)
 #if 0
 		/* set parameters, if needed */
 		if(SET_GEOM(Fs.Direct, &used_dev, 0xf0, boot)){
-			sprintf(errmsg,"Can't set disk parameters: %s", 
+			sprintf(errmsg,"Can't set disk parameters: %s",
 				strerror(errno));
 			continue;
 		}
@@ -1048,10 +1100,10 @@ void mformat(int argc, char **argv, int dummy)
 
 		SET_INT(Fs.sector_size, msize);
 		{
-		    unsigned int i;
-		    for(i = 0; i < 31; i++) {
-			if (Fs.sector_size == (unsigned int) (1 << i)) {
-			    Fs.sectorShift = i;
+		    unsigned int j;
+		    for(j = 0; j < 31; j++) {
+			if (Fs.sector_size == (unsigned int) (1 << j)) {
+			    Fs.sectorShift = j;
 			    break;
 			}
 		    }
@@ -1068,9 +1120,9 @@ void mformat(int argc, char **argv, int dummy)
 
 		/* do a "test" read */
 		if (!create &&
-		    READS(Fs.Direct, (char *) buf, 0, Fs.sector_size) != 
+		    READS(Fs.Direct, &boot.characters, 0, Fs.sector_size) !=
 		    (signed int) Fs.sector_size) {
-			sprintf(errmsg, 
+			sprintf(errmsg,
 				"Error reading from '%s', wrong parameters?",
 				name);
 			continue;
@@ -1086,31 +1138,43 @@ void mformat(int argc, char **argv, int dummy)
 		exit(1);
 	}
 
+	/* calculate the total number of sectors */
+	tot_sectors = used_dev.tracks*used_dev.heads*used_dev.sectors - used_dev.hidden;
+
+	/* create the image file if needed */
+	if (create) {
+		WRITES(Fs.Direct, &boot.characters,
+		       sectorsToBytes((Stream_t*)&Fs, tot_sectors-1),
+		       Fs.sector_size);
+	}
+
 	/* the boot sector */
-	boot = (struct bootsector *) buf;
 	if(bootSector) {
 		int fd;
 
-		fd = open(bootSector, O_RDONLY | O_LARGEFILE);
+		fd = open(bootSector, O_RDONLY | O_BINARY | O_LARGEFILE);
 		if(fd < 0) {
 			perror("open boot sector");
 			exit(1);
 		}
-		read(fd, buf, blocksize);
+		if(read(fd, &boot.bytes, blocksize) < blocksize) {
+			perror("short read on boot sector");
+			exit(1);
+		}
 		keepBoot = 1;
 	}
 	if(!keepBoot && !(used_dev.use_2m & 0x7f)) {
-		memset((char *)boot, '\0', Fs.sector_size);
+		memset(boot.characters, '\0', Fs.sector_size);
 		if(Fs.sector_size == 512 && !used_dev.partition) {
 			/* install fake partition table pointing to itself */
 			struct partition *partTable=(struct partition *)
-				(((char*) boot) + 0x1ae);
+				(&boot.bytes[0x1ae]);
 			setBeginEnd(&partTable[1], 0,
 						used_dev.heads * used_dev.sectors * used_dev.tracks,
 						used_dev.heads, used_dev.sectors, 1, 0);
 		}
 	}
-	set_dword(boot->nhs, used_dev.hidden);
+	set_dword(boot.boot.nhs, used_dev.hidden);
 
 	Fs.Next = buf_init(Fs.Direct,
 			   blocksize * used_dev.heads * used_dev.sectors,
@@ -1118,16 +1182,13 @@ void mformat(int argc, char **argv, int dummy)
 			   blocksize);
 	Fs.Buffer = 0;
 
-	boot->nfat = Fs.num_fat;
+	boot.boot.nfat = Fs.num_fat;
 	if(!keepBoot)
-		set_word(boot->jump + 510, 0xaa55);
+		set_word(&boot.bytes[510], 0xaa55);
 	
-	/* get the parameters */
-	tot_sectors = used_dev.tracks * used_dev.heads * used_dev.sectors - 
-		DWORD(nhs);
-
-	set_word(boot->nsect, used_dev.sectors);
-	set_word(boot->nheads, used_dev.heads);
+	/* Initialize the remaining parameters */
+	set_word(boot.boot.nsect, used_dev.sectors);
+	set_word(boot.boot.nheads, used_dev.heads);
 
 	used_dev.fat_bits = comp_fat_bits(&Fs,used_dev.fat_bits, tot_sectors, fat32);
 
@@ -1135,37 +1196,42 @@ void mformat(int argc, char **argv, int dummy)
 		Fs.primaryFat = 0;
 		Fs.writeAllFats = 1;
 		Fs.fat_start = 32;
-		calc_fs_parameters_32(tot_sectors, &Fs, boot);
+		calc_fs_parameters_32(tot_sectors, &Fs, &boot);
 
 		Fs.clus_start = Fs.num_fat * Fs.fat_len + Fs.fat_start;
 
 		/* extension flags: mirror fats, and use #0 as primary */
-		set_word(boot->ext.fat32.extFlags,0);
+		set_word(boot.boot.ext.fat32.extFlags,0);
 
 		/* fs version.  What should go here? */
-		set_word(boot->ext.fat32.fsVersion,fsVersion);
+		set_word(boot.boot.ext.fat32.fsVersion,fsVersion);
 
 		/* root directory */
-		set_dword(boot->ext.fat32.rootCluster, Fs.rootCluster = 2);
+		set_dword(boot.boot.ext.fat32.rootCluster, Fs.rootCluster = 2);
 
 		/* info sector */
-		set_word(boot->ext.fat32.infoSector, Fs.infoSectorLoc = 1);
+		set_word(boot.boot.ext.fat32.infoSector, Fs.infoSectorLoc = 1);
 		Fs.infoSectorLoc = 1;
 
 		/* no backup boot sector */
-		set_word(boot->ext.fat32.backupBoot, 6);
+		set_word(boot.boot.ext.fat32.backupBoot, 6);
 		
-		labelBlock = & boot->ext.fat32.labelBlock;
+		labelBlock = & boot.boot.ext.fat32.labelBlock;
 	} else {
 		Fs.infoSectorLoc = 0;
 		Fs.fat_start = 1;
-		calc_fs_parameters(&used_dev, tot_sectors, &Fs, boot);
+		calc_fs_parameters(&used_dev, tot_sectors, &Fs, &boot);
 		Fs.dir_start = Fs.num_fat * Fs.fat_len + Fs.fat_start;
 		Fs.clus_start = Fs.dir_start + Fs.dir_len;
-		labelBlock = & boot->ext.old.labelBlock;
+		labelBlock = & boot.boot.ext.old.labelBlock;
 
 	}
-	
+
+	/* Set the codepage */
+	Fs.cp = cp_open(used_dev.codepage);
+	if(Fs.cp == NULL)
+		exit(1);
+
 	if (!keepBoot)
 		/* only zero out physdrive if we don't have a template
 		 * bootsector */
@@ -1178,74 +1244,68 @@ void mformat(int argc, char **argv, int dummy)
 	if (!serial_set)
 		serial=random();
 	set_dword(labelBlock->serial, serial);	
-	if(!label[0])
-		strncpy(shortlabel, "NO NAME    ",11);
-	else
-		label_name(label, 0, &mangled, shortlabel);
-	strncpy(labelBlock->label, shortlabel, 11);
+	label_name(GET_DOSCONVERT((Stream_t *)&Fs),
+		   label[0] ? label : "NO NAME    ", 0, &mangled, &shortlabel);
+	strncpy(labelBlock->label, shortlabel.base, 11);
 	sprintf(labelBlock->fat_type, "FAT%2.2d  ", Fs.fat_bits);
 	labelBlock->fat_type[7] = ' ';
 
-	set_word(boot->secsiz, Fs.sector_size);
-	boot->clsiz = (unsigned char) Fs.cluster_size;
-	set_word(boot->nrsvsect, Fs.fat_start);
+	set_word(boot.boot.secsiz, Fs.sector_size);
+	boot.boot.clsiz = (unsigned char) Fs.cluster_size;
+	set_word(boot.boot.nrsvsect, Fs.fat_start);
 
-	bootOffset = init_geometry_boot(boot, &used_dev, sectors0, 
+	bootOffset = init_geometry_boot(&boot, &used_dev, sectors0,
 					rate_0, rate_any,
 					&tot_sectors, keepBoot);
 	if(!bootOffset) {
-		bootOffset = ((char *) labelBlock) - ((char *) boot) +
+		bootOffset = ((unsigned char *) labelBlock) - boot.bytes +
 			sizeof(struct label_blk_t);
 	}
 	if(Atari) {
-		boot->banner[4] = 0;
-		boot->banner[5] = random();
-		boot->banner[6] = random();
-		boot->banner[7] = random();
+		boot.boot.banner[4] = 0;
+		boot.boot.banner[5] = random();
+		boot.boot.banner[6] = random();
+		boot.boot.banner[7] = random();
 	}		
 
-	if (create) {
-		WRITES(Fs.Direct, (char *) buf,
-		       sectorsToBytes((Stream_t*)&Fs, tot_sectors-1),
-		       Fs.sector_size);
-	}
-
 	if(!keepBoot)
-		inst_boot_prg(boot, bootOffset);
+		inst_boot_prg(&boot, bootOffset);
 	/* Mimic 3.8 behavior, else 2m disk do not work (???)
 	 * luferbu@fluidsignal.com (Luis Bustamante), Fri, 14 Jun 2002
 	 */
 	if(used_dev.use_2m & 0x7f) {
-	  boot->jump[0] = 0xeb;
-	  boot->jump[1] = 0x80;
-	  boot->jump[2] = 0x90;
+	  boot.boot.jump[0] = 0xeb;
+	  boot.boot.jump[1] = 0x80;
+	  boot.boot.jump[2] = 0x90;
 	}
 	if(used_dev.use_2m & 0x7f)
 		Fs.num_fat = 1;
 	if(mediaDesc != -1)
-		boot->descr=mediaDesc;
+		boot.boot.descr=mediaDesc;
 	Fs.lastFatSectorNr = 0;
 	Fs.lastFatSectorData = 0;
-	zero_fat(&Fs, boot->descr);
+	zero_fat(&Fs, boot.boot.descr);
 	Fs.freeSpace = Fs.num_clus;
 	Fs.last = 2;
 
 #ifdef USE_XDF
 	if(format_xdf)
-		for(i=0; 
-		    i < (info.BadSectors+Fs.cluster_size-1)/Fs.cluster_size; 
+		for(i=0;
+		    i < (info.BadSectors+Fs.cluster_size-1)/Fs.cluster_size;
 		    i++)
 			fatEncode(&Fs, i+2, 0xfff7);
 #endif
 
-	format_root(&Fs, label, boot);
-	WRITES((Stream_t *)&Fs, (char *) boot, (mt_off_t) 0, Fs.sector_size);
-	if(Fs.fat_bits == 32 && WORD(ext.fat32.backupBoot) != MAX16) {
-		WRITES((Stream_t *)&Fs, (char *) boot, 
-		       sectorsToBytes((Stream_t*)&Fs, WORD(ext.fat32.backupBoot)),
+	format_root(&Fs, label, &boot);
+	WRITES((Stream_t *)&Fs, boot.characters,
+	       (mt_off_t) 0, Fs.sector_size);
+	if(Fs.fat_bits == 32 && WORD_S(ext.fat32.backupBoot) != MAX16) {
+		WRITES((Stream_t *)&Fs, boot.characters,
+		       sectorsToBytes((Stream_t*)&Fs,
+				      WORD_S(ext.fat32.backupBoot)),
 		       Fs.sector_size);
 	}
-	FLUSH((Stream_t *)&Fs); /* flushes Fs. 
+	FLUSH((Stream_t *)&Fs); /* flushes Fs.
 				 * This triggers the writing of the FAT */
 	FREE(&Fs.Next);
 	Fs.Class->freeFunc((Stream_t *)&Fs);
